package be.jeffcheasey88.peeratcode.parser.java.operations;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import be.jeffcheasey88.peeratcode.parser.java.CleanerPool;
import be.jeffcheasey88.peeratcode.parser.java.CleanerPool.Cleaner;
import be.jeffcheasey88.peeratcode.parser.java.JavaElement;
import be.jeffcheasey88.peeratcode.parser.java.Variable;
import be.jeffcheasey88.peeratcode.parser.java.operations.ConditionalOperation.ForOperation;
import be.jeffcheasey88.peeratcode.parser.java.operations.ConditionalOperation.IfOperation;
import be.jeffcheasey88.peeratcode.parser.java.operations.ConditionalOperation.WhileOperation;
import be.jeffcheasey88.peeratcode.parser.java.operations.LoopAffectOperation.BreakOperation;
import be.jeffcheasey88.peeratcode.parser.java.operations.LoopAffectOperation.ContinueOperation;
import be.jeffcheasey88.peeratcode.parser.java.operations.TryCatchOperation.CatchOperation;
import be.jeffcheasey88.peeratcode.parser.java.operations.TryCatchOperation.FinallyOperation;
import be.jeffcheasey88.peeratcode.parser.java.operations.TryCatchOperation.TryOperation;

public class OperationFactory{
	
	private static final OperationFactory SINGLETON = new OperationFactory();
	
	public static OperationFactory getFactory(){
		return SINGLETON;
	}
	
	private Map<Pattern, Class<? extends JavaElement>> patterns;
	
	private OperationFactory(){
		this.patterns = new LinkedHashMap<>();

		this.patterns.put(Pattern.compile("^\\s*return\\s*[^;]*;"), ReturnOperation.class);
		this.patterns.put(Pattern.compile("^\\s*if\\s*\\^GENERIC_PARENTHESIS\\d+"), IfOperation.class);
		this.patterns.put(Pattern.compile("^\\s*for\\s*\\^GENERIC_PARENTHESIS\\d+"), ForOperation.class);
		this.patterns.put(Pattern.compile("^\\s*while\\s*\\^GENERIC_PARENTHESIS\\d+"), WhileOperation.class);
		this.patterns.put(Pattern.compile("^\\s*else\\s*"), ElseOperation.class);
		this.patterns.put(Pattern.compile("^\\s*do\\s*\\^GENERIC_FUNCTION\\d+"), DoOperation.class);
		
		this.patterns.put(Pattern.compile("^\\s*synchronized\\s*\\^GENERIC_PARENTHESIS\\d+"), SynchronizedOperation.class);
		
		this.patterns.put(Pattern.compile("^\\s*continue\\s*;"), ContinueOperation.class);
		this.patterns.put(Pattern.compile("^\\s*break\\s*;"), BreakOperation.class);
		
		this.patterns.put(Pattern.compile("^\\s*throw\\s+[^;]+"), ThrowOperation.class);
		
		this.patterns.put(Pattern.compile("^\\s*try\\s*\\^"), TryOperation.class);
		this.patterns.put(Pattern.compile("^\\s*catch\\s*\\^GENERIC_PARENTHESIS\\d+"), CatchOperation.class);
		this.patterns.put(Pattern.compile("^\\s*finally\\s*\\^GENERIC_FUNCTION\\d+"), FinallyOperation.class);
		
		this.patterns.put(Pattern.compile("^\\s*[^\\^\\s]+\\^GENERIC_PARENTHESIS\\d+;"), MethodCallOperation.class);
		
		this.patterns.put(Pattern.compile("^\\s*[^\\s]+\\s+[^\\s]+\\s*=\\s*[^;]+;"), Variable.class);
		this.patterns.put(Pattern.compile("^\\s*[^\\s]+\\s*=\\s*[^;]+;"), AssigmentOperation.class);
		this.patterns.put(Pattern.compile("^\\s*[^;]+;"), Variable.class);
	}
	
	/*
	 * variable										OK
	 * assignation									OK
	 * exec method									OK
	 * for											OK
	 * do while										OK
	 * while										OK
	 * if											OK
	 * switch case
	 * return										OK
	 * continue										OK
	 * break										OK
	 * try catch finally							OK
	 * throw										OK
	 * else											OK
	 * synchronized									OK
	 * instance of
	 * 
	 * native operation style i++
	 * 
	 * assert ??
	 * Goto ??
	 * Const ??
	 */

	public JavaElement buildOperation(String content) throws Exception{
//		System.out.println("Factory.buildOperation -> "+content);
		CleanerPool generic = new CleanerPool(
				new Cleaner("GENERIC_ARRAY",'[',']'),
				new Cleaner("GENERIC_TYPE_",'<','>'));
		content = generic.clean(content);
		content = generiqueTypes(content, generic);

		for(Entry<Pattern, Class<? extends JavaElement>> entries : this.patterns.entrySet()){
			if(entries.getKey().matcher(content).lookingAt()) return entries.getValue().newInstance();
		}
		return null;
	}
	
	
	private static Pattern UNZIP_STICK = Pattern.compile("\\s+(?<e>[<|(|\\[|\"|'])");
	private static Pattern UNZIP_MAJ = Pattern.compile(">(?<e>[^>\\d,;(])");
	private static Pattern UNZIP_ARRAY = Pattern.compile("](?<e>[^\\[\\d,;])");
	private static Pattern UNZIP_EQUALS_LEFT = Pattern.compile("(?<e>[^=\\s])=");
	private static Pattern UNZIP_EQUALS_RIGHT = Pattern.compile("=(?<e>[^=\\s])");
	
	private String generiqueTypes(String content, CleanerPool cleaner){
		String unzip = cleaner.unzip(content, (value, pattern) -> {
				if(pattern.equals("^GENERIC_FUNCTION") || pattern.equals("^GENERIC_PARENTHESIS")) return null;
				return value.replaceAll("\\s+", "");
			});
		unzip = UNZIP_STICK.matcher(unzip).replaceAll("${e}");
		unzip = UNZIP_MAJ.matcher(unzip).replaceAll("> ${e}");
		unzip = UNZIP_ARRAY.matcher(unzip).replaceAll("] ${e}");
		unzip = UNZIP_EQUALS_LEFT.matcher(unzip).replaceAll("${e} =");
		unzip = UNZIP_EQUALS_RIGHT.matcher(unzip).replaceAll("= ${e}");
		return unzip;
	}
}
